# -*- coding: utf-8 -*-
# Copyright (c) 2017 the Residue team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""CRUD service for SQLAlchemy."""

from __future__ import absolute_import
import sys
import json
import collections
from copy import deepcopy
from functools import wraps

import six
from pockets import is_listy, listify, mappify
from pockets.autolog import log
from sqlalchemy import select, func
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.properties import ColumnProperty
from sqlalchemy.orm.util import class_mapper
from sqlalchemy.sql import ClauseElement
from sqlalchemy.sql.expression import cast, and_, or_, asc, desc, literal
from sqlalchemy.types import Text, Integer, String

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache  # noqa: F401


__all__ = ['normalize_sort', 'normalize_data', 'normalize_query', 'crud_exceptions', 'make_crud_api', 'CrudException']


class CrudException(Exception):
    pass


def _collect_fields(d):
    if 'field' in d:
        return {d['field']}
    elif 'and' in d or 'or' in d:
        attrs = set()
        for comp in ['and', 'or']:
            for subquery in d.get(comp, []):
                attrs.update(_collect_fields(subquery))
        return attrs
    elif 'comparison' in d or 'value' in d:
        return {'id'}
    else:
        return d.keys()


def _extract_sort_field(model, value, index=0):
    field = None
    fields = listify(value)
    for f in fields:
        if isinstance(f, six.string_types):
            parts = f.split('.')
            if len(parts) == 1 and field is None:
                if not model or (model and hasattr(model, parts[0])):
                    field = parts[0]
            elif len(parts) > 1 and model and parts[0] == model.__name__:
                field = parts[1]
        else:
            field = f

    if field and isinstance(field, six.string_types) and model:
        attr = getattr(model, field)
        if not (isinstance(attr, InstrumentedAttribute) and isinstance(attr.property, ColumnProperty)) \
                and not isinstance(attr, ClauseElement):
            raise ValueError(
                'SQLAlchemy model classes may only be sorted by columns that exist in the database. '
                'Provided: {}.{}'.format(model.__name__, field))
    return field or 'id'


def _get_queries(x):
    queries = []
    if isinstance(x, (list, tuple)):
        for e in x:
            queries.extend(_get_queries(e))
    elif isinstance(x, dict):
        queries.append(x)
        for e in x.values():
            queries.extend(_get_queries(e))
    return [d for d in queries if isinstance(d.get("_model"), six.string_types)]


def normalize_sort(model, sort):
    if sort and isinstance(sort, six.string_types):
        first_char = sort.lstrip()[0]
        if first_char == '[' or first_char == '{':
            sort = json.loads(sort)

    if isinstance(sort, six.string_types):
        return [{'field': _extract_sort_field(model, sort), 'dir': 'asc'}]
    elif is_listy(sort):
        sorters = []
        for s in sort:
            sorters.extend(normalize_sort(model, s))
        return sorters
    elif isinstance(sort, dict):
        field = sort.get('property', sort.get('fields', sort.get('field', [])))
        direction = sort.get('direction', sort.get('dir', 'asc')).lower()
        return [{'field': _extract_sort_field(model, field), 'dir': direction}]
    return [{'field': 'id', 'dir': 'asc'}]


def normalize_data(data, count=1):
    """
    Normalizes a CrudApi data parameter.

    A singular data can be a string, a list of strings, or a dict::

        'attr'
        ['attr1', 'attr2']
        {'attr1':True, 'attr2':True}


    A plural data must be specified as a list of lists or a list of dicts::

        [['attr1', 'attr2'], ['attr1', 'attr2']]
        [{'attr1':True, 'attr2':True}, {'attr1':True, 'attr2':True}]


    Note that if data is specified as a list of strings, it is
    considered to be singular. Only a list of lists or a list of
    dicts is considered plural.

    Returns the plural form of data as the comprehensive form of a list of
    dictionaries mapping <keyname> to True, extended to count length. If a
    singular data is given, the result will be padded by repeating
    that value. If a plural data is given, it will be padded with
    None, for example::

        >>> normalize_data('attr', 1)
        [{'attr': True}]
        >>> normalize_data('attr', 3)
        [{'attr': True}, {'attr': True}, {'attr': True}]

    """
    if not data:
        return listify(None, minlen=count)

    if isinstance(data, six.string_types):
        data = [{data: True}]
    elif isinstance(data, collections.Mapping):
        data = [data]
    elif isinstance(data, collections.Iterable):
        if any(isinstance(element, six.string_types) for element in data):
            data = [data]
        data = [mappify(v) for v in data]
    else:
        raise TypeError('Unknown datatype: {}: {!r}', type(data), data)

    if len(data) < count:
        if len(data) == 1:
            data.extend([deepcopy(data[0]) for i in range(count - len(data))])
        else:
            data.extend([None for i in range(count - len(data))])
    return data


def normalize_query(query, top_level=True, supermodel=None):
    """
    Normalizes a variety of query formats to a known standard query format.

    The comprehensive form of the query parameter is as follows::

        query = [{
            '_model': 'ModelClassName',
            '_label': 'Optional identifier',

            # Either provide <logical_operator> OR <comparison>
            <logical_operator>|<comparison>
        }]+

        logical_operator =
            'OR'|'AND': [<query>[, <query>]*]

        comparison =
            'comparison': <comparison_function>,
            'field': 'model_field_name',
            'value': 'model_field_value'

        comparison_function =
            'eq'|'ne'|'lt'|'le'|'gt'|'ge'|'in'|'notin'|'isnull'|'isnotnull'|
            'contains'|'icontains'|'like'|'ilike'|'startswith'|'endswith'|
            'istartswith'|'iendswith'

    """
    if query is None:
        raise ValueError('None passed for query parameter')

    query = listify(deepcopy(query))

    queries = []
    for q in query:
        if isinstance(q, six.string_types):
            queries.append({'_model': q, '_label': q})
        elif isinstance(q, dict):
            if 'distinct' in q:
                if isinstance(q['distinct'], six.string_types):
                    q['distinct'] = [q['distinct']]

            if 'groupby' in q:
                if isinstance(q['groupby'], six.string_types):
                    q['groupby'] = [q['groupby']]

            if 'and' in q or 'or' in q:
                op = 'and' if 'and' in q else 'or'

                if not isinstance(q[op], (list, set, tuple)):
                    raise ValueError('Clause must be of type list, set, or tuple not {}, given {}'.format(
                        type(q[op]), q[op]))

                q[op] = normalize_query(q[op], False, q.get('_model', supermodel))
                if len(q[op]) == 1:
                    q = q[op][0]
                elif '_model' not in q:
                    # Pull the _model up from the sub clauses. Technically the
                    # query format requires the _model be declared in the
                    # clause, but we are going to be liberal in what we accept.
                    model = supermodel
                    for clause in q[op]:
                        if '_model' in clause:
                            model = clause['_model']
                            break

                    if model is None:
                        raise ValueError('Clause objects must have a "_model" attribute, given:\n{}'.format(q))
                    q['_model'] = model

            if '_model' in q:
                queries.append(q)
            elif supermodel is not None:
                q['_model'] = supermodel
                queries.append(q)
            else:
                raise ValueError('Query objects must have a "_model" attribute, given:\n{}'.format(q))
        else:
            raise ValueError('Query objects must be either a dict or string, given {}:\n{}'.format(type(q), q))
    return queries


def crud_exceptions(fn):
    """A decorator designed to catch exceptions from the crud api methods."""
    @wraps(fn)
    def wrapped(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception:
            a = [x for x in (args or [])]
            kw = {k: v for k, v in (kwargs or {}).items()}
            log.error('Error calling {}.{} {!r} {!r}'.format(fn.__module__, fn.__name__, a, kw), exc_info=True)
            exc_class, exc, tb = sys.exc_info()
            raise six.reraise(CrudException, CrudException(str(exc)), tb)
    return wrapped


def make_crud_api(Session):

    class CrudApi(object):

        @classmethod
        def _collect_models(cls, query):
            models = set()
            for d in listify(query):
                try:
                    model = Session.resolve_model(d['_model'])
                except Exception:
                    log.debug('unable to resolve model {} in query {}', d.get('_model'), d)
                else:
                    models.add(model)
                    for attr_name in _collect_fields(d):
                        curr_model = model
                        for prop_name in attr_name.split('.'):
                            if hasattr(curr_model, prop_name):
                                prop = getattr(curr_model, prop_name)
                                if isinstance(prop, InstrumentedAttribute) and hasattr(prop.property, 'mapper'):
                                    curr_model = prop.property.mapper.class_
                                    models.update([curr_model])
                                    if prop_name in d:
                                        subquery = deepcopy(d[prop_name])
                                        if isinstance(subquery, (list, set, tuple)) \
                                                and not filter(lambda x: isinstance(x, dict), subquery):
                                            subquery = {i: True for i in subquery}
                                        elif isinstance(subquery, six.string_types):
                                            subquery = {subquery: True}
                                        if isinstance(subquery, dict):
                                            subquery['_model'] = curr_model.__name__
                                        models.update(cls._collect_models(subquery))
                            else:
                                break
            return models

        @classmethod
        def _get_models(cls, *args, **kwargs):
            return {model.__name__ for model in cls._collect_models(_get_queries([args, kwargs]))}

        @classmethod
        def _sort_query(cls, query, model, sort):
            sort = normalize_sort(model, sort)
            for sorter in sort:
                dir = {'asc': asc, 'desc': desc}[sorter['dir']]
                field = sorter['field']
                if model:
                    field = getattr(model, field)
                    if issubclass(type(field.__clause_element__().type), String):
                        field = func.lower(field)
                query = query.order_by(dir(field))
            return query

        @classmethod
        def _limit_query(cls, query, limit, offset):
            if offset is not None:
                query = query.offset(offset)
            if limit is not None and limit != 0:
                query = query.limit(limit)
            return query

        # this only works in postgresql
        @classmethod
        def _distinct_query(cls, query, filters):
            distinct_clause = filters.get('distinct', None)
            if distinct_clause:
                if isinstance(distinct_clause, bool):
                    query = query.distinct()
                else:
                    model = Session.resolve_model(filters.get('_model'))
                    columns = [getattr(model, field) for field in distinct_clause]
                    query = query.distinct(*columns)
            return query

        @classmethod
        def _groupby_query(cls, query, filters):
            groupby_clause = filters.get('groupby', None)
            if groupby_clause:
                model = Session.resolve_model(filters.get('_model'))
                columns = [getattr(model, field) for field in groupby_clause]
                query = query.group_by(*columns)
            return query

        @classmethod
        def _filter_query(cls, query, model, filters=None, limit=None, offset=None, sort=None):
            if filters:
                query = cls._distinct_query(query, filters)
                query = cls._groupby_query(query, filters)
                filters = cls._resolve_filters(filters, model)
                if filters is not None:
                    query = query.filter(filters)
            if sort:
                query = cls._sort_query(query, model, sort)
            query = cls._limit_query(query, limit, offset)
            return query

        @classmethod
        def _resolve_comparison(cls, comparison, column, value):
            if isinstance(value, dict):
                model_class = Session.resolve_model(value.get('_model'))
                field = value.get('select', 'id')
                value = select([getattr(model_class, field)], cls._resolve_filters(value))

            return {
                'eq': lambda field, val: field == val,
                'ne': lambda field, val: field != val,
                'lt': lambda field, val: field < val,
                'le': lambda field, val: field <= val,
                'gt': lambda field, val: field > val,
                'ge': lambda field, val: field >= val,
                'in': lambda field, val: field.in_(val),
                'notin': lambda field, val: ~field.in_(val),
                'isnull': lambda field, val: field == None,
                'isnotnull': lambda field, val: field != None,
                'contains': lambda field, val: field.like('%'+val+'%'),
                'icontains': lambda field, val: field.ilike('%'+val+'%'),
                'like': lambda field, val: field.like('%'+val+'%'),
                'ilike': lambda field, val: field.ilike('%'+val+'%'),
                'startswith': lambda field, val: field.startswith(val),
                'endswith': lambda field, val: field.endswith(val),
                'istartswith': lambda field, val: field.ilike(val+'%'),
                'iendswith': lambda field, val: field.ilike('%'+val)
            }[comparison](column, value)  # noqa: E711

        @classmethod
        def _resolve_filters(cls, filters, model=None):
            model = Session.resolve_model(filters.get('_model', model))
            and_clauses = filters.get('and', None)
            or_clauses = filters.get('or', None)
            if and_clauses:
                return and_(*[cls._resolve_filters(c, model) for c in and_clauses])
            elif or_clauses:
                return or_(*[cls._resolve_filters(c, model) for c in or_clauses])
            elif 'field' in filters or 'value' in filters or 'comparison' in filters:
                field = filters.get('field', 'id').split('.')
                value = filters.get('value')
                comparison = filters.get('comparison', 'eq')

                if len(field) == 1:
                    column = getattr(model, field[0])
                    return cls._resolve_comparison(comparison, column, value)
                elif len(field) == 2:
                    property = field[0]
                    field = field[1]
                    related_table = class_mapper(model).get_property(property)
                    related_model = related_table.argument
                    if isinstance(related_model, Mapper):
                        related_model = related_model.class_
                    elif callable(related_model):
                        related_model = related_model()
                    related_field = getattr(related_model, field)

                    clause = cls._resolve_comparison(comparison, related_field, value)
                    if getattr(related_table, 'primaryjoin', None) is not None:
                        clause = and_(
                            clause,
                            related_table.primaryjoin)
                    if getattr(related_table, 'secondaryjoin', None) is not None:
                        clause = and_(
                            clause,
                            related_table.secondaryjoin)
                    return clause
                else:
                    property = field[0]
                    join_property = field[1]
                    field = field[2]

                    join_table = class_mapper(model).get_property(property)
                    join_model = join_table.argument

                    if isinstance(join_model, Mapper):
                        join_model = join_model.class_
                    elif callable(join_model):
                        join_model = join_model()

                    related_table = class_mapper(join_model).get_property(join_property)
                    related_model = related_table.argument
                    if isinstance(related_model, Mapper):
                        related_model = related_model.class_
                    elif callable(related_model):
                        related_model = related_model()
                    related_field = getattr(related_model, field)

                    clause = cls._resolve_comparison(comparison, related_field, value)
                    if getattr(join_table, 'primaryjoin', None) is not None:
                        clause = and_(
                            clause,
                            join_table.primaryjoin)
                    if getattr(join_table, 'secondaryjoin', None) is not None:
                        clause = and_(
                            clause,
                            join_table.secondaryjoin)

                    if getattr(related_table, 'primaryjoin', None) is not None:
                        clause = and_(
                            clause,
                            related_table.primaryjoin)
                    if getattr(related_table, 'secondaryjoin', None) is not None:
                        clause = and_(
                            clause,
                            related_table.secondaryjoin)

                    return clause
            else:
                return None

        @crud_exceptions
        def count(self, query):
            """
            Count the model objects matching the supplied query parameters

            @param query: Specifies the model types to count. May be a string, a list
            of strings, or a list of dicts with a "_model" key specified.

            Returns:
                int: The count of each of the supplied model types, in a list of
                    dicts, like so::

                        [{
                            '_model' : 'Player',
                            '_label' : 'Player on a Team',
                            'count' : 12
                        }]

            """
            filters = normalize_query(query)
            results = []
            with Session() as session:
                for filter in filters:
                    model = Session.resolve_model(filter['_model'])
                    result = {'_model': filter['_model'],
                              '_label': filter.get('_label', filter['_model'])}
                    if getattr(model, '_crud_perms', {}).get('read', True):
                        if filter.get('groupby', False):
                            columns = []
                            for attr in filter['groupby']:
                                columns.append(getattr(model, attr))

                            rows = CrudApi._filter_query(
                                session.query(func.count(columns[0]), *columns), model, filter).all()
                            result['count'] = []
                            for row in rows:
                                count = {'count': row[0]}
                                index = 1
                                for attr in filter['groupby']:
                                    count[attr] = row[index]
                                    index += 1
                                result['count'].append(count)
                        else:
                            result['count'] = CrudApi._filter_query(session.query(model), model, filter).count()
                    results.append(result)
            return results

        @crud_exceptions
        def read(self, query, data=None, order=None, limit=None, offset=0):
            """
            Get the model objects matching the supplied query parameters,
            optionally setting which part of the objects are in the returned dictionary
            using the supplied data parameter

            Arguments:
                session (sqlalchemy.Session): SQLAlchemy session to use.
                query: one or more queries (as c{dict} or [c{dict}]), corresponding
                    to the format of the query parameter described in the module-level
                    docstrings. This query parameter will be normalized
                data: one or more data specification (as c{dict} or [c{dict}]),
                    corresponding to the format of the data specification parameter
                    described in the module-level docstrings. The length of the data
                    parameter should either be 1 which will be the spec for each query
                    specified, OR of length N, where N is the number of queries after
                    normalization. If not provided the _data parameter will be expected
                    in each query
                limit: The limit parameter, when provided with positive integer "L"
                    at most "L" results will be returned. Defaults to no limit
                offset: The offset parameter, when provided with positive integer
                    "F", at most "L" results will be returned after skipping the first "F"
                    results (first based on ordering)

            Returns:
                list: one or more data specification dictionaries with models that
                    match the provided queries including all readable fields without
                    following foreign keys (the default if no data parameter is included),
                    OR the key/values specified by the data specification parameter. The
                    number of items returned and the order in which they appear are
                    controlled by the limit, offset and order parameters. Represented as::

                        return {
                            total: <int> # count of ALL matching objects, separate from <limit>
                            results: [c{dict}, c{dict}, ... , c{dict}] # subject to <limit>
                        }

            """
            with Session() as session:
                filters = normalize_query(query)
                data = normalize_data(data, len(filters))
                if len(filters) == 1:
                    filter = filters[0]
                    model = Session.resolve_model(filter['_model'])
                    total = 0
                    results = []
                    if getattr(model, '_crud_perms', {}).get('read', True):
                        total = CrudApi._filter_query(session.query(model), model, filter).count()
                        results = CrudApi._filter_query(session.query(model), model, filter, limit, offset, order).all()

                    return {'total': total, 'results': [r.crud_read(data[0]) for r in results]}

                elif len(filters) > 1:
                    queries = []
                    count_queries = []
                    queried_models = []
                    sort_field_types = {}
                    for filter_index, filter in enumerate(filters):
                        model = Session.resolve_model(filter['_model'])
                        if getattr(model, '_crud_perms', {}).get('read', True):
                            queried_models.append(model)
                            query_fields = [
                                model.id,
                                cast(literal(model.__name__), Text).label("_table_name"),
                                cast(literal(filter_index), Integer)]

                            for sort_index, sort in enumerate(normalize_sort(model, order)):
                                sort_field = getattr(model, sort['field'])
                                sort_field_types[sort_index] = type(sort_field.__clause_element__().type)
                                query_fields.append(sort_field.label('anon_sort_{}'.format(sort_index)))
                            queries.append(CrudApi._filter_query(session.query(*query_fields), model, filter))
                            count_queries.append(CrudApi._filter_query(session.query(model.id), model, filter))

                    total = count_queries[0].union(*(count_queries[1:])).count()
                    query = queries[0].union(*(queries[1:]))
                    normalized_sort_fields = normalize_sort(None, order)
                    for sort_index, sort in enumerate(normalized_sort_fields):
                        dir = {'asc': asc, 'desc': desc}[sort['dir']]
                        sort_field = 'anon_sort_{}'.format(sort_index)
                        if issubclass(sort_field_types[sort_index], String):
                            sort_field = 'lower({})'.format(sort_field)
                        query = query.order_by(dir(sort_field))
                    if normalized_sort_fields:
                        query = query.order_by("_table_name")
                    rows = CrudApi._limit_query(query, limit, offset).all()

                    result_table = {}
                    result_order = {}
                    query_index_table = {}
                    for i, row in enumerate(rows):
                        id = str(row[0])
                        model = Session.resolve_model(row[1])
                        query_index = row[2]
                        result_table.setdefault(model, []).append(id)
                        result_order[id] = i
                        query_index_table[id] = query_index

                    for model, ids in result_table.items():
                        result_table[model] = session.query(model).filter(model.id.in_(ids)).all()

                    ordered_results = len(result_order) * [None]
                    for model, instances in result_table.items():
                        for instance in instances:
                            ordered_results[result_order[instance.id]] = instance
                    results = [r for r in ordered_results if r is not None]

                    return {'total': total, 'results': [r.crud_read(data[query_index_table[r.id]]) for r in results]}
                else:
                    return {'total': 0, 'results': []}

        @crud_exceptions
        def create(self, data):
            """
            Create a model object using the provided data specifications.

            Arguments:
                session (sqlalchemy.Session): SQLAlchemy session to use.
                data (dict or [dict]): One or more residue data dictionaries. A
                    new object will be created for each data specification
                    dictionary provided.

            Returns:
                list: A list containing the newly created objects.

            """
            data = normalize_data(data)
            if any('_model' not in attrs for attrs in data):
                raise CrudException('_model is required to create a new item')

            created = []
            with Session() as session:
                for attrs in data:
                    model = Session.resolve_model(attrs['_model'])
                    instance = model()
                    session.add(instance)
                    instance.crud_create(**attrs)
                    session.flush()  # any items that were created should now be queryable
                    created.append(instance.crud_read())
            return created

        @crud_exceptions
        def update(self, query, data):
            """
            Updates the model objects matching the supplied query parameter.

            The matching objects will be updated according to the fields and values
            specified in the data parameter.

            Arguments:
                session (sqlalchemy.Session): SQLAlchemy session to use.
                query (dict or [dict]): One or more residue query dictionaries.
                    This query parameter will be normalized.
                data (dict or [dict]): One or more residue data dictionaries. The
                    length of the data parameter should be N, where N is the number
                    of queries after normalization.

            Returns:
                bool: True if the objects were successfully updated.

            """
            filters = normalize_query(query)
            data = normalize_data(data, len(filters))
            with Session() as session:
                for filter, attrs in zip(filters, data):
                    model = Session.resolve_model(filter['_model'])
                    for instance in CrudApi._filter_query(session.query(model), model, filter):
                        instance.crud_update(**attrs)
                        # any items that were created should now be queryable
                        session.flush()
            return True

        @crud_exceptions
        def delete(self, query):
            """
            Delete the model objects matching the supplied query parameters

            Arguments:
                session (sqlalchemy.Session): SQLAlchemy session to use.
                query (dict or [dict]): One or more residue query dictionaries.
                    This query parameter will be normalized.

            Returns:
                int: The number of objects successfully deleted.

            """
            deleted = 0
            filters = normalize_query(query)
            with Session() as session:
                for filter in filters:
                    model = Session.resolve_model(filter['_model'])
                    if getattr(model, '_crud_perms', {}).get('can_delete', False):
                        to_delete = CrudApi._filter_query(session.query(model), model, filter)
                        count = to_delete.count()
                        assert count in [0, 1], "each query passed to crud.delete must return at most 1 item"
                        if count == 1:
                            # don't log if there wasn't actually a deletion
                            item_to_delete = to_delete.one()
                            session.delete(item_to_delete)
                            deleted += count
            return deleted

    return CrudApi()



# -*- coding: utf-8 -*-
# Copyright (c) 2017 the Residue team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""SQLAlchemy CRUD Utilities"""

from __future__ import absolute_import
import inspect
import re
import uuid
from types import MethodType

import six
import sqlalchemy
from pockets import camel, uncamel, collect_subclasses
from pockets.autolog import log
from sqlalchemy.ext import declarative
from sqlalchemy.orm import configure_mappers, sessionmaker, Query

from residue._version import __version__  # noqa: F401
from residue.crud.api import make_crud_api
from residue.crud.orm import CrudModelMixin

from residue.crud import *  # noqa: F401,F403
from residue.query import *  # noqa: F401,F403
from residue.types import *  # noqa: F401,F403


NAMESPACE_SQL = uuid.UUID('75c7e3be-a5c7-414d-bc66-d64ae5d03f3d')
RE_NONCHAR_WHITESPACE = re.compile('[\W\s]+')


def check_constraint_naming_convention(constraint, table):
    """
    Creates a unique name for an unnamed CheckConstraint.

    The generated name is the SQL text of the CheckConstraint with
    non-alphanumeric, non-underscore operators converted to text, and all
    other non-alphanumeric, non-underscore substrings replaced by underscores.

    If the generated name is longer than 32 characters, a uuid5 based on the
    generated name will be returned instead.

    >>> from sqlalchemy import CheckConstraint, MetaData, Table
    >>> table = Table('account', MetaData())
    >>> constraint = CheckConstraint('failed_logins > 3')

    See: http://docs.sqlalchemy.org/en/latest/core/constraints.html#configuring-constraint-naming-conventions

    """
    # The text of the replacements doesn't matter, so long as it's unique
    replacements = [
        ('||/', 'cr'), ('<=', 'le'), ('>=', 'ge'), ('<>', 'nq'), ('!=', 'ne'),
        ('||', 'ct'), ('<<', 'ls'), ('>>', 'rs'), ('!!', 'fa'), ('|/', 'sr'),
        ('@>', 'cn'), ('<@', 'cb'), ('&&', 'an'), ('<', 'lt'), ('=', 'eq'),
        ('>', 'gt'), ('!', 'ex'), ('"', 'qt'), ('#', 'hs'), ('$', 'dl'),
        ('%', 'pc'), ('&', 'am'), ('\'', 'ap'), ('(', 'lpr'), (')', 'rpr'),
        ('*', 'as'), ('+', 'pl'), (',', 'cm'), ('-', 'da'), ('.', 'pd'),
        ('/', 'sl'), (':', 'co'), (';', 'sc'), ('?', 'qn'), ('@', 'at'),
        ('[', 'lbk'), ('\\', 'bs'), (']', 'rbk'), ('^', 'ca'), ('`', 'tk'),
        ('{', 'lbc'), ('|', 'pi'), ('}', 'rbc'), ('~', 'td')]

    constraint_name = str(constraint.sqltext).strip()
    for operator, text in replacements:
        constraint_name = constraint_name.replace(operator, text)

    constraint_name = RE_NONCHAR_WHITESPACE.sub('_', constraint_name)
    if len(constraint_name) > 32:
        constraint_name = uuid.uuid5(NAMESPACE_SQL, str(constraint_name)).hex
    return constraint_name


# Consistent naming conventions are necessary for alembic to be able to
# reliably upgrade and downgrade versions. For more details, see:
# http://alembic.zzzcomputing.com/en/latest/naming.html
default_naming_convention = {
    'unnamed_ck': check_constraint_naming_convention,
    'ix': 'ix_%(column_0_label)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(unnamed_ck)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'pk': 'pk_%(table_name)s'}


# SQLAlchemy doesn't expose its default constructor as a nicely importable
# function, so we grab it from the function defaults.
_spec_args, _spec_varargs, _spec_kwargs, _spec_defaults = inspect.getargspec(declarative.declarative_base)
declarative_base_constructor = dict(zip(reversed(_spec_args), reversed(_spec_defaults)))['constructor']


def declarative_base(*orig_args, **orig_kwargs):
    """
    Replacement for SQLAlchemy's declarative_base, which adds these features:
    1) This is a decorator.
    2) This allows your base class to set a constructor.
    3) This provides a default constructor which automatically sets defaults
       instead of waiting to do that until the object is committed.
    4) Automatically setting __tablename__ to snake-case.
    5) Automatic integration with the SessionManager class.
    """
    orig_args = list(orig_args)

    def _decorate_base_class(klass):

        class Mixed(klass, CrudModelMixin):
            def __init__(self, *args, **kwargs):
                """
                Variant on SQLAlchemy model __init__ which sets default values on
                initialization instead of immediately before the model is saved.
                """
                if '_model' in kwargs:
                    assert kwargs.pop('_model') == self.__class__.__name__
                declarative_base_constructor(self, *args, **kwargs)
                for attr, col in self.__table__.columns.items():
                    if kwargs.get(attr) is None and col.default:
                        self.__dict__.setdefault(attr, col.default.execute())

        orig_kwargs['cls'] = Mixed
        if 'name' not in orig_kwargs:
            orig_kwargs['name'] = klass.__name__
        if 'constructor' not in orig_kwargs:
            orig_kwargs['constructor'] = klass.__init__ if '__init__' in klass.__dict__ else Mixed.__init__

        Mixed = declarative.declarative_base(*orig_args, **orig_kwargs)
        Mixed.BaseClass = _SessionInitializer._base_classes[klass.__module__] = Mixed
        Mixed.__tablename__ = declarative.declared_attr(lambda cls: uncamel(cls.__name__))
        return Mixed

    is_class_decorator = not orig_kwargs \
        and len(orig_args) == 1 \
        and inspect.isclass(orig_args[0]) \
        and not isinstance(orig_args[0], sqlalchemy.engine.Connectable)

    if is_class_decorator:
        return _decorate_base_class(orig_args.pop())
    else:
        return _decorate_base_class


class _SessionInitializer(type):
    _base_classes = {}

    def __new__(cls, name, bases, attrs):
        SessionClass = type.__new__(cls, name, bases, attrs)
        if hasattr(SessionClass, 'engine'):
            if not hasattr(SessionClass, 'BaseClass'):
                for module, bc in _SessionInitializer._base_classes.items():
                    if module == SessionClass.__module__:
                        SessionClass.BaseClass = bc
                        break
                else:
                    raise AssertionError('No BaseClass specified and @declarative_base was never invoked in {}'.format(
                        SessionClass.__module__))
            if not hasattr(SessionClass, 'session_factory'):
                SessionClass.session_factory = sessionmaker(
                    bind=SessionClass.engine, autoflush=False, autocommit=False, query_cls=SessionClass.QuerySubclass)

            SessionClass.initialize_db()
            SessionClass.crud = make_crud_api(SessionClass)
        return SessionClass


@six.add_metaclass(_SessionInitializer)
class SessionManager(object):
    class SessionMixin(object):
        pass

    class QuerySubclass(Query):
        pass

    def __init__(self):
        self.session = self.session_factory()
        for name, val in self.SessionMixin.__dict__.items():
            if not name.startswith('__'):
                assert not hasattr(self.session, name) and hasattr(val, '__call__')
                setattr(self.session, name, MethodType(val, self.session))

    def __enter__(self):
        return self.session

    def __exit__(self, exc_type, exc_value, traceback):
        try:
            if exc_type is None:
                self.session.commit()
        finally:
            self.session.close()

    def __del__(self):
        if self.session.transaction._connections:
            log.error('SessionManager went out of scope without underlying connection being closed; '
                      'did you forget to use it as a context manager?')
            self.session.close()

    @classmethod
    def initialize_db(cls, drop=False, create=True):
        configure_mappers()
        cls.BaseClass.metadata.bind = cls.engine
        if drop:
            cls.BaseClass.metadata.drop_all(cls.engine, checkfirst=True)
        if create:
            cls.BaseClass.metadata.create_all(cls.engine, checkfirst=True)

    @classmethod
    def all_models(cls):
        return collect_subclasses(cls.BaseClass)

    @classmethod
    def resolve_model(cls, name):
        if inspect.isclass(name) and issubclass(name, cls.BaseClass):
            return name

        subclasses = {ModelClass.__name__: ModelClass for ModelClass in cls.all_models()}
        permutations = [name, camel(name), camel(name, upper_segments=0)]
        for name in permutations:
            if name in subclasses:
                return subclasses[name]

            if name.lower().endswith('s'):
                singular = name.rstrip('sS')
                if singular in subclasses:
                    return subclasses[singular]

            if name.lower().endswith('ies'):
                singular = name[:-3] + 'y'
                if singular in subclasses:
                    return subclasses[singular]

        for name in permutations:
            if name in cls.BaseClass.metadata.tables:
                return cls.BaseClass.metadata.tables[name]

        raise ValueError('Unrecognized model: {}'.format(name))



import json
import uuid

import six
from pytz import UTC
from sqlalchemy.dialects import postgresql
from sqlalchemy.types import CHAR, DateTime, String, TypeDecorator, Unicode


__all__ = ['CoerceUTF8', 'UUID', 'JSON', 'UTCDateTime']


class CoerceUTF8(TypeDecorator):
    """
    Safely coerce Python bytestrings to unicode before sending to the database.
    """
    impl = Unicode

    def process_bind_param(self, value, dialect):
        if isinstance(value, type(b'')):
            value = value.decode('utf-8')
        return value


class UUID(TypeDecorator):
    """
    Platform-independent UUID type.

    Uses Postgresql's UUID type if available. Otherwise stores as a hex
    formatted CHAR(32).

    """
    impl = CHAR

    def load_dialect_impl(self, dialect):
        if dialect.name == 'postgresql':
            return dialect.type_descriptor(postgresql.UUID())
        else:
            return dialect.type_descriptor(CHAR(32))

    def process_bind_param(self, value, dialect):
        if value is None:
            return value
        elif dialect.name == 'postgresql':
            return str(value)
        else:
            if not isinstance(value, uuid.UUID):
                return uuid.UUID(value).hex
            else:
                return value.hex

    def process_result_value(self, value, dialect):
        if value is None:
            return value
        else:
            return str(uuid.UUID(value))


class JSON(TypeDecorator):
    impl = String

    def __init__(self, comparator=None):
        self.comparator = comparator
        super(JSON, self).__init__()

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        elif isinstance(value, six.string_types):
            return value
        else:
            return json.dumps(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return json.loads(str(value))

    def copy_value(self, value):
        if self.mutable:
            return json.loads(json.dumps(value))
        else:
            return value

    def compare_values(self, x, y):
        if self.comparator:
            return self.comparator(x, y)
        else:
            return x == y


class UTCDateTime(TypeDecorator):
    impl = DateTime

    def process_bind_param(self, value, engine):
        if value is not None:
            return value.astimezone(UTC).replace(tzinfo=None)

    def process_result_value(self, value, engine):
        if value is not None:
            return value.replace(tzinfo=UTC)



"""
The `crud` module defines a number of functions for finding SQLAlchemy model
objects via a query parameter and displaying a desired portion of the resulting
object graph via a data specification parameter, optionally limiting the total
number returned, potentially with an offset to support paging.


QUERY PARAMETER
---------------
The format of the query parameter needs to support logical operators and a
certain amount of introspection into which model objects are involved in a
give query. For this writeup, a "query" is any set of search parameters that
will result in a known SQL search string capable of returning the desired
model objects. Python syntax will be used to represent the expected format of
the method parameters, with allowances for representing infinite nesting/lists
as appropriate. Unless explicitly stated, pluralized forms like "queries" can
be read as "query or queries" due to the support of one or more queries in all
cases.

The comprehensive form of the query parameter is as follows::

    query = [{
        '_model': <model_name>,
        '_data': <data_specification>,
        '_label': <query_label>,
        # Either provide <logical_operator> OR items after <logical_operator>
        <logical_operator>: [<query>[, <query>]*],
        # Used IF AND ONLY IF <logical_operator> is not provided
        'comparison': <comparison_function>
        'field': <model_field_name>,
        'value': <model_field_value>
    }]

meaning an array of one or more dictionaries (a dictionary is equivalent to an
array of length 1) of queries, one for each type of SQLAlchemy model object
expected to be returned.

where:
- '<model_name>' - the string corresponding to the SQLAlchemy model class name
    which extends your @residue.declarative_base
- '<query_label>' - the optional string that signifies the purpose of this
    query and is only used as a convenience for the consumer of the crud
    method. This primarily supports counts, but can used in client code to
    help cue the display of those results, defaults to the contents of _model
- '<logical_operator>' - the key is one of the following logical operators
    (with the value being one of more queries in a list):
-- and ("intersection")
-- or ("union")
--- meaning that the results of the provided queries will be the corresponding
    intersection/union of all the results of an individual query. Imagining a
    Venn Diagram is useful in this instance.
- <query> - is a dictionary identical the dictionary taken in by the query
    parameter EXCEPT that _model is not included
- <comparison> - a comparison operator function used to find the objects that
    would return "True" for the provided comparison for the value in the
    model_field_name. Some examples are:
-- 'lt' - is the field less than value?
-- 'gt' - is the field greater than value?
-- 'eq' - is the field equal to value? (default)
-- 'ne' - is the field not equal value?
-- 'le' - is the field less than or equal to value?
-- 'ge' - is the field greater than or equal to value?
-- 'isnull' - does the field have a null value? (does not use the query's
    value parameter)
-- 'isnotnull' - does the field have a non null value? (does not use the
    query's value parameter)
-- 'in' - does the field appear in the given value? (value should be an array)
-- 'contains' - does the field contain this value? (would allow other
    characters before or after the value)
-- 'like' - same as contains, but case sensitive
-- 'ilike' - same as contains, case insensitive
-- 'startswith' - does the field start with value?
-- 'istartswith' - case insensitive startswith
-- 'endswith' - does the field end with value?
-- 'iendswith' - case insensitive endswith

- <model_field_name> - the name of the field for the provided _model at the
    top level. Supports dot notation, e.g.:
-- making a comparison based off all of a Team's players' names would use an
    'field' of 'player.name'
- <model_field_value> - the value that the field comparison will be made
    against, e.g. a value of 'text' and a comparison of 'eq' will return all
    matching models with fields equal to 'text'.
- <data_specification> - specifying what parts of the results get returned,
    the following section covers the format the data specification parameter


DATA SPECIFICATION
------------------
Where the query parameter is only used to search existing objects in the
database, the data specification parameter has two separate meanings: in the
'read' function as the _data key in the query dictionary: what information is
returned in the results, in the 'update' and 'create' functions, what model
type will be created/updated with what values. This is encompassed in one
format, so there is some amount of redundancy depending on what actions you're
performing.

The comprehensive form of the data specification parameter is as follows:

data = [{
    '_model': <model_name>,
    # a non-foreign key field
    '<model_field_name>': True (or the value of the field if the data
        parameter is used to create or update objects)
    # a foreign key field is a special case and additional forms are supported)
    '<foreign_key_model_field_name>': True (all readable fields of the
        reference model object will be read. Has no meaning if the data
        parameter is used to create of update objects)
    '<foreign_key_model_field_name>': {<same form as the data parameter, e.g.,
        supports recursion}
}] +

Meaning an array of one or more dictionaries (a dictionary is equivalent to an
array of length 1) of data specification, one for each type of LDAP model
object expected to be returned. As a special case for the 'read' method, one
dictionary is interpreted as being the intended data spec for each item in the
query parameter array. In the 'update' array, the length of the query and data
parameters must match and the nth member of both the query and data array are
read together as a matched set. In the 'create' method, each member of the
data array will create a new object of the type specified in _model.

A supported short form of a data specification is, instead of a dictionary of
key names with values, a list of key names that should be read:

['<model_field_name_1',
 '<model_field_name_2',
 '<model_field_name_3']

is equivalent to:

{'<model_field_name_1': True,
 '<model_field_name_2': True,
 '<model_field_name_3': True}

As you can see, that this short form would not be appropriate for create or
updates function calls, as there's no way to specify the desired values.
Additionally there's no way to specify a sub-object graph for a followed
foreign key.


RESULTS FOR crud.count
----------------------
The crud.count method accepts a query parameter (format examined above) and
returns a count of each of the supplied queries (typically, this is a count of
each supplied model type), however the results also include a _label key, that
can be used to differentiate between two different types of results within the
same model type (e.g. enabled accounts vs disabled accounts)

e.g.:

return [{
    _model : 'Team',
    _label: 'Team',
    count : 12
}, {
    _model : 'Player',
    _label : 'Players on a Team',
    count : 144
}, {
    _model : 'Manager',
    _label : 'Managers of a Team',
    count : 12
}, {
    _model : 'LeagueEmployee',
    _label : 'Everyone employed by the league (e.g. Players, Managers)',
    count : 156
}]


RESULTS FOR crud.read
---------------------
The crud.read method accepts both a query and data specification parameter
(format examined above), and two parameters for fine-tuning which specific
results are returned (examined in the upcoming "Fine-Tuning Read Results"
section. The read method returns the total number of objects matching the
query (separate from any sort of limits) and a list of the specific objects
requested (subject to those limits) e.g.

return {
    total: 20 # count of ALL matching objects
    # although only 5 results were returned as a result of the
    # specified fine-tuning parameters
    results: [<result>, <result>, <result>, <result>, <result>]
}

To prevent the client from always being forced to deal with entire query
result, there are three parameters in place for the crud.read method to
simplify only receiving the information that's desired. At a high level:

- 'limit' takes a positive integer 'L' and when provided, the crud.read
    method will return at most L results, defaults to no limit
- 'order' takes a list of order specification dictionaries for sorting by
    specific fields and in a specified direction (ascending or descending),
    defaults to no ordering
- 'offset' takes a positive integer 'F' and when provided, the crud.read
    method will return at most L results, after skipping the first (based
    on the order specification) F results.

Used with the crud.read method to only return only a subset of information,
allowing the client to only receive the amount of information it's interested
in. Useful in conjunction with the offset and ordering parameter to
finely-tune the information received.

The comprehensive form of the order parameter is as follows:

order = [{
    'dir': <'asc'/'desc'> # either in ascending (default) or descending order
    'fields': [['<model_object_name>.]<model_field_name>']+
}] +

A single string in 'fields' is equivalent to a list with the string as the
only element. If no model_object name is provided, the model_field_name is
interpreted as the catch-all key for all model objects. If model_field_name
isn't present on a model, or no catch-all is specified, 'id' will be used

The list of dictionaries are interpreted as being ordered in decreasing
priority. An example:

The 'offset' parameter is used with the crud.read method to only return only
a subset of information, allowing the client to only receive the amount of
information it's interested in. Useful in conjunction with the limit and
ordering parameter to finely-tune the information received.

Using the 4 records in the order example (including the order specification):
- a limit of 1 with an offset of 0 (the default if unspecified) would return
    only the John Depp Human.
- a limit of 0 (unlimited, which is the default if unspecified) and an offset
    of 0 would be identical to the table in the order-only example
- a limit of 0 and an offset of 1 would return everything except for the first
    result, so in this case, the last 3 results
- a limit of 2 and an offset of 1 would return the 2nd and 3rd results, so in
    this case, the middle 2 results

"""


from residue.crud.api import *  # noqa: F401,F403
from residue.crud.orm import *  # noqa: F401,F403



# -*- coding: utf-8 -*-
# Copyright (c) 2017 the Residue team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""SQLAlchemy query functions."""

from __future__ import absolute_import
import re
from datetime import datetime

from sqlalchemy.sql import and_, bindparam, literal, func, select, text, union


__all__ = ['constrain_query_by_date', 'generate_date_series', 'normalize_query_by_date']


RE_STARTSWITH_DIGIT = re.compile(r'^\s*\d+')


def constrain_query_by_date(query, column, start_date=None, end_date=None, interval='1 month'):
    """
    Applies a WHERE clause which constrains a query within start and end dates.

    For example, if column is "user.birthdate" and query looks like::

        select name from user;

    The result would look something like this::

        select name from user
        where birthdate > '2017-01-01' and birthdate < '2017-01-02';

    Args:
        query (sqlalchemy.orm.query.Query): The query to which the date
            constraining WHERE clause should be applied.
        column (sqlalchemy.sql.schema.Column): The date column which should be
            used in the constraining WHERE clause.
        start_date (datetime): The earliest date in the constraint.
        end_date (datetime): The most recent date in the constraint.
        interval (str): Alternately, the length of time from either
            `start_date` or `end_date` expressed as a Postgres interval.
            Defaults to '1 month'.

    Returns:
        sqlalchemy.orm.query.Query: The constrained query.
    """
    if start_date:
        if end_date:
            # If the start_date and the end_date are defined then we use those
            return query.where(and_(column >= start_date, column <= end_date))
        elif interval:
            # If the start_date and the interval are defined then we use the
            # start_date plus the interval as the end_date
            return query.where(and_(
                column >= start_date,
                column <= text('DATE :start_date_param_1 + INTERVAL :interval_param_1', bindparams=[
                    bindparam('start_date_param_1', start_date),
                    bindparam('interval_param_1', interval)])))
        else:
            # If ONLY the start_date is defined then we just use that
            return query.where(column >= start_date)
    elif end_date:
        if interval:
            # If the end_date and the interval are defined then we use the
            # end_date minus the interval as the start_date
            return query.where(and_(
                column <= end_date,
                column >= text('DATE :end_date_param_1 - INTERVAL :interval_param_1', bindparams=[
                    bindparam('end_date_param_1', end_date),
                    bindparam('interval_param_1', interval)])))
        else:
            # If ONLY the end_date is defined then we just use that
            return query.where(column <= end_date)
    elif interval:
        # If ONLY the interval is defined then we use the current date minus
        # the interval as the start_date
        return query.where(column >= text('DATE :current_date_param_1 - INTERVAL :interval_param_1', bindparams=[
            bindparam('current_date_param_1', datetime.utcnow()),
            bindparam('interval_param_1', interval)]))

    # If NOTHING was defined then the query is returned unmodified
    return query


def generate_date_series(start_date=None, end_date=None, interval='1 month', granularity='1 day'):
    """
    Generates a date series; useful for grouping queries into date "buckets".

    The resulting query looks something like this::

        generate_series(DATE '2017-04-17', DATE '2017-05-07', '1 day')

    Args:
        start_date (datetime): Start date of the series.
        end_date (datetime): End date of the series.
        interval (str): Alternately, the length of time from either
            `start_date` or `end_date` expressed as a Postgres interval.
            Defaults to '1 month'.
        granularity (str): The granularity of each date "bucket" expressed
            as a Postgres interval. Defaults to '1 day'.

    Returns:
        sqlalchemy.orm.query.Query: A date series query.
    """
    if not granularity:
        granularity = '1 day'
    elif not RE_STARTSWITH_DIGIT.match(granularity):
        granularity = '1 {}'.format(granularity)

    if start_date:
        if end_date:
            # If the start_date and the end_date are defined then we use those
            return func.generate_series(start_date, end_date, granularity)
        elif interval:
            # If the start_date and the interval are defined then we use the
            # start_date plus the interval as the end_date
            return func.generate_series(
                start_date,
                text('DATE :start_date_param_1 + INTERVAL :interval_param_1', bindparams=[
                     bindparam('start_date_param_1', start_date),
                     bindparam('interval_param_1', interval)]),
                granularity)
        else:
            # If ONLY the start_date is defined then we use the current date
            # as the end_date
            return func.generate_series(start_date, datetime.utcnow(), granularity)

    if not end_date:
        # If the start_date and end_date are both undefined, we set the
        # end_date to the current date
        end_date = datetime.utcnow()

    if not interval:
        # If the interval is undefined, we set the interval to "1 month"
        interval = '1 month'

    return func.generate_series(
        text('DATE :end_date_param_1 - INTERVAL :interval_param_1', bindparams=[
            bindparam('end_date_param_1', end_date),
            bindparam('interval_param_1', interval)]),
        end_date,
        granularity)


def normalize_query_by_date(query, date_label, report_label, start_date=None, end_date=None, interval='1 month',
                            granularity='1 day'):
    """
    Fills in missing date "buckets" for an aggregate query grouped by date.

    Aggregate queries grouped by date are often used for generating reports,
    like "How many widgets did I sell on each day of last month?" These queries
    often look similar to this::

        SELECT date_of_sale AS day, count(id) AS sales_count
        FROM sales
        WHERE date_of_sale > now() - interval '1 month'
        GROUP BY day ORDER BY day;

    This kind of query will ONLY return days that contain at least one sale.
    Days with zero sales will not be returned at all, leaving gaps in the
    report. To combat this, we can normalize the query by generating a UNION
    with a Postgres date series::

        SELECT date_of_sale AS day, coalesce(sales_count, 0) AS sales_count
        FROM (
          SELECT date_of_sale AS day, count(id) AS sales_count
          FROM sales
          WHERE date_of_sale > now() - interval '1 month'
          GROUP BY day
        UNION
          SELECT
            generate_series(now() - interval '1 month', now(), '1 day') AS day,
            0 AS sales_count
        ) AS union_query ORDER BY union_query.day;

    Args:
        query (sqlalchemy.orm.query.Query): The original query grouped by date,
            in which the "group by date" column is labelled using
            column.label(date_label), and the "aggregate" column is labelled
            using column.label(report_label).
        date_label (str): The label used to label the "group by date" column.
        report_label (str): The label used to label the "aggregate" column.
        start_date (datetime): Start date of the query.
        end_date (datetime): End date of the query.
        interval (str): Alternately, the length of time from either
            `start_date` or `end_date` expressed as a Postgres interval.
            Defaults to '1 month'.
        granularity (str): The granularity of each date "bucket" expressed
            as a Postgres interval. Defaults to '1 day'.

    Returns:
        sqlalchemy.orm.query.Query: A normalized aggregate date query.
    """
    series = generate_date_series(start_date, end_date, interval, granularity)
    series_query = select([series.label(date_label), literal(0).label(report_label)])
    query = union(query, series_query).alias()
    query = select([
        text(date_label),
        func.coalesce(text(report_label), literal(0)).label(report_label)
    ], from_obj=query).order_by(date_label)
    return query



# -*- coding: utf-8 -*-
# Copyright (c) 2017 the Residue team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""SQLAlchemy CRUD Utilities"""

import os
from setuptools import setup, find_packages


# Package versioning solution originally found here:
# http://stackoverflow.com/q/458550
exec(open(os.path.join('residue', '_version.py')).read())

reqs = open('requirements.txt', 'r').read().strip().splitlines()

setup(
    name='residue',
    version=__version__,
    url='http://residue.readthedocs.org',
    download_url='http://pypi.python.org/pypi/residue',
    license='BSD',
    author='Rob Ruana',
    author_email='rob@robruana.com',
    description=__doc__,
    long_description=open('README.rst', 'r').read(),
    zip_safe=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
        'Topic :: Utilities',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    platforms='any',
    packages=find_packages(exclude=['tests*']),
    install_requires=reqs
)



# -*- coding: utf-8 -*-
# Copyright (c) 2017 the Residue team, see AUTHORS.
# Licensed under the BSD License, see LICENSE for details.

"""CRUD support for SQLAlchemy orm objects."""

from __future__ import absolute_import
import inspect
import re
import uuid
from collections import defaultdict, Callable, Mapping
from copy import deepcopy
from itertools import chain
from datetime import date, datetime, time

import six
from pockets import cached_classproperty, classproperty, collect_superclass_attr_names, is_data, is_listy, mappify
from pockets.autolog import log
from sqlalchemy import orm
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound
from sqlalchemy.orm.properties import ColumnProperty, RelationshipProperty
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy.sql import ClauseElement
from sqlalchemy.types import Boolean, DateTime, Integer, String, Text, UnicodeText

try:
    from functools import lru_cache
except ImportError:
    from backports.functools_lru_cache import lru_cache  # noqa: F401


__all__ = [
    'get_model_by_table', 'get_primary_key_column_names', 'get_unique_constraint_column_names',
    'get_one_to_many_foreign_key_column_name', 'CrudModelMixin', 'crudable', 'crud_validation',
    'text_length_validation', 'regex_validation']


@lru_cache()
def get_model_by_table(base, table):
    """
    Returns declarative class associated with given table.

    Arguments:
        base (sqlalchemy.ext.declarative.api.Base): Declarative model base or
            a subclass of the declarative model base.
        table (sqlalchemy.sql.schema.Table): SQLAlchemy Table object.

    Returns:
        class: Declarative class or None if not found.
    """
    for cls in base._decl_class_registry.values():
        if hasattr(cls, '__table__') and cls.__table__ is table:
            return cls
    return None


@lru_cache()
def get_one_to_many_foreign_key_column_name(model, name):
    """
    Returns the constituent column names for the foreign key on the remote
    table of the one-to-many relationship specified by name.

    Args:
        model (class or object): The given model class or model instance.
        name (string): The name of the attribute on `model` which is a
            one-to-many relationship.

    Return:
        list: One-to-many foreign key column names as a list of strings.
    """
    if not inspect.isclass(model):
        return get_one_to_many_foreign_key_column_name(model.__class__, name)

    attr = getattr(model, name, None)
    if not attr:
        # Unknown attribute.
        return []

    remote_columns = getattr(attr.property, 'remote_side', None)
    if not remote_columns:
        # This is not a one-to-many relationship.
        return []

    remote_tables = set(c.table.name for c in remote_columns)
    if len(remote_tables) > 1:
        # This is a many-to-many relationship with a cross reference table.
        return []

    foreign_key_column_names = []
    for remote_column in remote_columns:
        if getattr(remote_column, 'foreign_keys', False):
            foreign_key_column_names.append(remote_column.name)
        else:
            remote_model = get_model_by_table(model, remote_column.table)
            if remote_model:
                # Quasi foreign keys don't actually have foreign_keys set,
                # but they need to be treated as though they did.
                foreign_keys = getattr(remote_model, 'quasi_foreign_keys', [])
                if remote_column.name in foreign_keys:
                    foreign_key_column_names.append(remote_column.name)

    return foreign_key_column_names


@lru_cache()
def get_primary_key_column_names(model):
    """
    Returns the constituent column names for the primary key of the given
    model.

    Args:
        model (class or object): The given model class or model instance.

    Return:
        list: Primary key column names as a list of strings.
    """
    return [column.name for column in model.__table__.primary_key.columns]


@lru_cache()
def get_unique_constraint_column_names(model):
    """
    Returns the constituent column names for each unique constraint on the
    given model.

    Args:
        model (class or object): The given model class or model instance.

    Return:
        list: Unique constraint column names as a list of lists of strings.
    """
    return [[column.name for column in constraint.columns]
            for constraint in model.__table__.constraints
            if isinstance(constraint, UniqueConstraint)]


class CrudModelMixin(object):
    extra_defaults = []
    type_casts = {uuid.UUID: str}
    type_map = {}
    type_map_defaults = {
        int: 'int',
        six.binary_type: 'string',
        six.text_type: 'string',
        float: 'float',
        datetime: 'date',
        date: 'date',
        time: 'date',
        bool: 'boolean',
        uuid.UUID: 'string',
        String: 'string',
        UnicodeText: 'string',
        Text: 'string',
        DateTime: 'date',
        Integer: 'int',
        Boolean: 'boolean',
    }

    # Override what attributes will show in the repr. Defaults to primary keys
    # and unique constraints.
    repr_attr_names = ()

    # In addition to any default attributes, also show these in the repr.
    extra_repr_attr_names = ()

    # columns that are used as foreign keys, but don't have an explicit
    # foreign key constraint
    quasi_foreign_keys = ()

    @classmethod
    def _create_or_fetch(cls, session, value, **backref_mapping):
        """
        Fetch an existing or create a new instance of this class. Fetching uses
        the values from the value positional argument (the id if available, or
        if any keys that correspond to unique constraints are present). In both
        cases the instance will still need to be updated using whatever new
        values you want.

        Args:
            cls (class): The class object we're going to fetch or create
            session (Session): the session object
            value (any): the dictionary value to fetch with
            **backref_mapping: the backref key name and value of the "parent"
                object of the object you're fetching or about to create. If
                the backref value of a fetched instance is not the same as the
                value of what's passed in, we will instead create a new
                instance. This is because we want to prevent "stealing" an
                existing object in a one-to-one relationship unless an id is
                explicitly passed.

        Returns:
            A previously existing or newly created (and added to the session)
            model instance.
        """
        assert len(backref_mapping) <= 1, 'only one backref key is allowed at this time: {}'.format(backref_mapping)

        if backref_mapping:
            backref_name = list(backref_mapping.keys())[0]
            parent_id = backref_mapping[backref_name]
        else:
            backref_name, parent_id = None, None

        id = None
        if isinstance(value, Mapping):
            id = value.get('id', None)
        elif isinstance(value, six.string_types):
            id = value

        instance = None
        if id is not None:
            try:
                instance = session.query(cls).filter(cls.id == id).first()
            except Exception:
                log.error('Unable to fetch instance based on id value {!r}', value, exc_info=True)
                raise TypeError('Invalid instance ID type for relation: {} (value: {})'.format(cls.__name__, value))
        elif isinstance(value, Mapping):
            # If there's no id, check to see if we have a dictionary that
            # includes all of the columns associated with a UniqueConstraint.
            for column_names in cls.unique_constraint_column_names:
                if all((s in value and value[s]) for s in column_names):
                    # all those column names are provided,
                    # use that to query by chaining together all the necessary
                    # filters to construct that query
                    q = session.query(cls)
                    filter_kwargs = dict((s, value[s]) for s in column_names)
                    try:
                        instance = q.filter_by(**filter_kwargs).one()
                    except NoResultFound:
                        continue
                    except MultipleResultsFound:
                        log.error('Multiple results found for {} unique constraint: {}', cls.__name__, column_names)
                        raise
                    else:
                        break
                else:
                    log.debug('Unable to search using unique constraints: {} with {}', column_names, value)

        if instance and id is None and backref_mapping and getattr(instance, backref_name, None) != parent_id:
            log.warning(
                'Attempting to change the owner of {} without an explicitly '
                'passed id; a new {} instance will be used instead',
                instance, cls.__name__)
            instance = None

        if not instance:
            log.debug('creating new: {} with id {}', cls.__name__, id)
            if id is None:
                instance = cls()
            else:
                instance = cls(id=id)
            session.add(instance)
        return instance

    @property
    @lru_cache()
    def _type_casts_for_to_dict(self):
        type_casts = CrudModelMixin.type_casts.copy()
        type_casts.update(self.type_casts)
        return defaultdict(lambda: lambda x: x, type_casts)

    @classproperty
    def to_dict_default_attrs(cls):
        attr_names = []
        super_attr_names = collect_superclass_attr_names(cls, terminal_class=cls.BaseClass)
        for name in super_attr_names:
            if not name.startswith('_') or name in cls.extra_defaults:
                attr = getattr(cls, name)
                descriptor = getattr(attr, 'descriptor', None)

                is_attr = not (
                    callable(attr)
                    or isinstance(descriptor, hybrid_property)
                    or isinstance(attr, (property, InstrumentedAttribute, ClauseElement)))
                is_instrumented = isinstance(attr, InstrumentedAttribute) and isinstance(attr.property, ColumnProperty)

                if is_attr or is_instrumented:
                    attr_names.append(name)
        return attr_names

    def to_dict(self, attrs=None, validator=lambda self, name: True):
        obj = {}
        if attrs is not None:
            attrs = mappify(attrs)

        # It's still possible for the client to blacklist this, but by default
        # we're going to include them.
        if attrs is None or attrs.get('_model', True):
            obj['_model'] = self.__class__.__name__
        if attrs is None or attrs.get('id', True):
            obj['id'] = self.id

        def cast_type(value):
            # Ensure that certain types are cast appropriately for daily usage
            # e.g. we want the result of HashedPasswords to be the string
            # representation instead of the object.
            return self._type_casts_for_to_dict[value.__class__](value)

        if attrs is None:
            for name in self.to_dict_default_attrs:
                if validator(self, name):
                    obj[name] = cast_type(getattr(self, name))
        else:
            for name in self.extra_defaults + list(attrs.keys()):
                # If we're not supposed to get the attribute according to the
                # validator, OR the client intentionally blacklisted it, then
                # skip this value.
                if not validator(self, name) or not attrs.get(name, True):
                    continue
                attr = getattr(self, name, None)
                if isinstance(attr, self.BaseClass):
                    obj[name] = attr.to_dict(attrs[name], validator)
                elif isinstance(attr, (list, set, tuple, frozenset)):
                    obj[name] = []
                    for item in attr:
                        if isinstance(item, self.BaseClass):
                            obj[name].append(
                                item.to_dict(attrs[name], validator))
                        else:
                            obj[name].append(item)
                elif callable(attr):
                    obj[name] = cast_type(attr())
                else:
                    obj[name] = cast_type(attr)

        return obj

    def from_dict(self, attrs, validator=lambda self, name, val: True):
        relations = []
        # Merge_relations modifies the dictionaries that are passed to it in
        # order to support updates in deeply-nested object graphs. To ensure
        # that we don't have dirty state between applying updates to different
        # model objects, we need a fresh copy.
        attrs = deepcopy(attrs)
        for name, value in attrs.items():
            if not name.startswith('_') and validator(self, name, value):
                attr = getattr(self.__class__, name)
                if isinstance(attr, InstrumentedAttribute) and isinstance(attr.property, RelationshipProperty):
                    relations.append((name, value))
                else:
                    setattr(self, name, value)

        def required(kv):
            cols = list(getattr(self.__class__, kv[0]).property.local_columns)
            return len(cols) != 1 or cols[0].primary_key or cols[0].nullable
        relations.sort(key=required)

        for name, value in relations:
            self._merge_relations(name, value, validator)

        return self

    @classproperty
    def primary_key_column_names(cls):
        return get_primary_key_column_names(cls)

    @classproperty
    def unique_constraint_column_names(cls):
        return get_unique_constraint_column_names(cls)

    @classmethod
    def one_to_many_foreign_key_column_name(cls, name):
        column_names = get_one_to_many_foreign_key_column_name(cls, name)
        return column_names[0] if column_names else None

    def _merge_relations(self, name, value, validator=lambda self, name, val: True):
        attr = getattr(self.__class__, name)
        if not isinstance(attr, InstrumentedAttribute) or not isinstance(attr.property, RelationshipProperty):
            return

        session = orm.Session.object_session(self)
        assert session, "Can't call _merge_relations on objects not attached to a session"

        attr_property = attr.property
        relation_cls = attr_property.mapper.class_

        # E.g., if this a Team with many Players, and we're handling the
        # attribute name "players," we want to set the team_id on all
        # dictionary representations of those players.
        backref_id_name = self.one_to_many_foreign_key_column_name(name)
        original_value = getattr(self, name)

        if is_listy(original_value):
            new_instances = []
            if value is None:
                value = []

            if isinstance(value, six.string_types):
                value = [value]

            for item in value:
                if backref_id_name is not None and isinstance(item, dict) and not item.get(backref_id_name):
                    item[backref_id_name] = self.id

                relation_inst = relation_cls._create_or_fetch(
                    session, item, **{backref_id_name: self.id} if backref_id_name else {})

                if isinstance(item, dict):
                    if relation_inst._sa_instance_state.identity:
                        validator = _crud_write_validator
                    else:
                        validator = _crud_create_validator
                    relation_inst.from_dict(item, validator)

                new_instances.append(relation_inst)

            relation = original_value
            remove_instances = [stale for stale in relation if stale not in new_instances]
            for stale_instance in remove_instances:
                relation.remove(stale_instance)
                if attr_property.cascade.delete_orphan:
                    session.delete(stale_instance)

            for new_instance in new_instances:
                if new_instance.id is None or new_instance not in relation:
                    relation.append(new_instance)

        elif isinstance(value, (Mapping, six.string_types)):
            if backref_id_name is not None and not value.get(backref_id_name):
                # If this is a dictionary, it's possible we're going to be
                # creating a new thing, if so, we'll add a backref to the
                # "parent" if one isn't already set.
                value[backref_id_name] = self.id

            relation_instance = relation_cls._create_or_fetch(session, value)
            stale_instance = original_value
            if stale_instance and stale_instance.id != relation_instance.id and attr_property.cascade.delete_orphan:
                session.delete(stale_instance)

            if isinstance(value, Mapping):
                relation_instance.from_dict(value, validator)
                # We want this this to be immediatelt queryable.
                session.flush([relation_instance])

            setattr(self, name, relation_instance)

        elif value is None:
            # The first branch handles the case of setting a many-to-one value
            # to None. So this is for the one-to-one-mapping case.
            # Setting a relation to None is nullifying the relationship, which
            # has potential side effects in the case of cascades, etc...
            setattr(self, name, value)
            stale_instance = original_value
            if stale_instance and attr_property.cascade.delete_orphan:
                session.delete(stale_instance)

        else:
            raise TypeError(
                'Merging relations on {} not supported for values of type: {} '
                '(value: {})'.format(name, value.__class__.__name__, value))

    def __setattr__(self, name, value):
        if name in getattr(self, '_validators', {}):
            for validator_dict in self._validators[name]:
                if not validator_dict['model_validator'](self, value):
                    message = validator_dict.get('validator_message')
                    raise ValueError(
                        'Validation failed for {}.{} with value {!r}: {}'
                        .format(self.__class__.__name__, name, value, message))
        object.__setattr__(self, name, value)

    def crud_read(self, attrs=None):
        return self.to_dict(attrs, validator=_crud_read_validator)

    def crud_create(self, **kwargs):
        return self.from_dict(kwargs, validator=_crud_create_validator)

    def crud_update(self, **kwargs):
        return self.from_dict(kwargs, validator=_crud_write_validator)

    def __repr__(self):
        """
        Useful string representation for logging.

        Note:
            __repr__ does NOT return unicode on Python 2, since python decodes
            it using the default encoding: http://bugs.python.org/issue5876.

        """
        # If no repr attr names have been set, default to the set of all
        # unique constraints. This is unordered normally, so we'll order and
        # use it here.
        if not self.repr_attr_names:
            # Flatten the unique constraint list
            _unique_attrs = chain(*self.unique_constraint_column_names)
            _primary_keys = self.primary_key_column_names

            attr_names = tuple(sorted(set(chain(_unique_attrs, _primary_keys, self.extra_repr_attr_names))))
        else:
            attr_names = self.repr_attr_names

        if not attr_names and hasattr(self, 'id'):
            # There should be SOMETHING, so use id as a fallback.
            attr_names = ('id',)

        if attr_names:
            _kwarg_list = ' '.join('{}={!r}'.format(s, getattr(self, s, 'undefined')) for s in attr_names)
            kwargs_output = ' {}'.format(_kwarg_list)
        else:
            kwargs_output = ''

        # Specifically using the string interpolation operator and the repr of
        # getattr so as to avoid any "hilarious" encode errors for non-ascii
        # characters.
        u_str = '<%s%s>' % (self.__class__.__name__, kwargs_output)
        return u_str if six.PY3 else u_str.encode('utf-8')


def _crud_read_validator(self, name):
    _crud_perms = getattr(self, '_crud_perms', None)
    if _crud_perms is not None and not _crud_perms.get('read', True):
        raise ValueError('Attempt to read non-readable model {}'.format(self.__class__.__name__))
    elif name in self.extra_defaults:
        return True
    elif _crud_perms is None:
        return not name.startswith('_')
    else:
        return name in _crud_perms.get('read', {})


def _crud_write_validator(self, name, value=None):
    _crud_perms = getattr(self, '_crud_perms', None)
    if getattr(self, name, None) == value:
        return True
    elif not _crud_perms or not _crud_perms.get('update', False):
        raise ValueError('Attempt to update non-updateable model {}'.format(self.__class__.__name__))
    elif name not in _crud_perms.get('update', {}):
        raise ValueError('Attempt to update non-updateable attribute {}.{}'.format(self.__class__.__name__, name))
    else:
        return name in _crud_perms.get("update", {})


def _crud_create_validator(self, name, value=None):
    _crud_perms = getattr(self, '_crud_perms', {})
    if not _crud_perms or not _crud_perms.get('can_create', False):
        raise ValueError('Attempt to create non-createable model {}'.format(self.__class__.__name__))
    else:
        return name in _crud_perms.get("create", {})


class crudable(object):
    """
    Decorator that specifies which model attributes are part of the CRUD API.

    Intended to be used on SQLAlchemy model classes, for example::

        @crudable(
            create=True,
            read=['__something'],
            no_read=['password'],
            update=[],
            no_update=[],
            delete=True,
            data_spec={
                attr={
                    read=True,
                    update=True,
                    desc='description'
                    defaultValue=<some default>
                    validators={<validator_name>, <validator value>}
                }
            }
        )
        class MyModelObject(Base):
            # ...

    The resulting object will have a class attribute named "crud_spec" which
    is a dictionary like::

        {
            create: True/False,
            read: {<attribute name>, <attribute name>},
            update: {<attribute name>, <attribute name>},
            delete: True/False,
            data_spec: {
                manually_specified_attr: {
                    desc: 'description',
                    type: '<type>'
                    read: True/False # only needed if attribute is unspecified
                    update": True/False
                }
                attr_with_manual_description: {
                    desc: 'description',
                    type: '<type>'
                }
            }
        }

    Attributes:
        never_read (tuple): Names of attributes that default to being not
            readable.
        never_update (tuple): Names of attribute that default to being not
            updatable.
        always_create (tuple): Names of attributes that default to being always
            creatable.
        default_labels (dict): Attribute name and label pairs, to simplify
            setting the same label for each and every instance of an attribute
            name.
    """

    never_read = ('metadata',)
    never_update = ('id',)
    always_create = ('id',)
    default_labels = {}

    def __init__(
            self,
            can_create=True,
            create=None,
            no_create=None,
            read=None,
            no_read=None,
            update=None,
            no_update=None,
            can_delete=True,
            data_spec=None):
        """
        Args:
            can_create (bool): If True (default), the decorated class can be
                created.
            create (collections.Iterable): If provided, interpreted as the
                attribute names that can be specified when the object is
                created in addition to the items are updateable. If not
                provided (default) all attributes that can be updated plus the
                primary key are allowed to be passed to the create method.
            no_create (collections.Iterable): If provided, interpreted as the
                attribute names that will not be allowed to be passed to
                create, taking precedence over anything specified in the create
                parameter. If not provided (default) everything allowed by the
                create parameter will be acceptable.
            read (collections.Iterable): If provided, interpreted as the
                attribute names that can be read, and ONLY these names can be
                read. If not provided (default) all attributes not starting
                with an underscore (e.g. __str__, or _hidden) will be readable,
            no_read (collections.Iterable): if provided, interpreted as the
                attribute names that can't be read, taking precedence over
                anything specified in the read parameter. If not provided
                (default) everything allowed by the read parameter will be
                readable.
            update (collections.Iterable): If provided, interpreted as the
                attribute names that can be updated, in addition to the list of
                items are readable. If None (default) default to the list of
                readable attributes. Pass an empty iterable to use the default
                behavior listed under the read docstring if there were
                attributes passed to read that you don't want update to default
                to.
            no_update (collections.Iterable): if provided, interpreted as the
                attribute names that can't be updated, taking precedence over
                anything specified in the update parameter. If None (default)
                default to the list of non-readable attributes. Pass an empty
                iterable to use the default behavior listed under the no_read
                docstring if there were attributes passed to no_read that you
                don't want no_update to default to.
            can_delete (bool): If True (default), the decorated class can be
                deleted.
            data_spec (dict): Any additional information that should be added
                to the `model.get_crud_definition`. See that function for
                complete documentation, but the key items are:
                "desc" - Human-readable description, will default to docstrings
                    if available, else not be present in the final spec.
                "label" - a Human-readable short label to help remember the
                    purpose of a particular field, without going into detail.
                    If not specifically provided, it will not be present in the
                    spec.
                "type" - the human-readable "type" for an attribute meaning
                    that a conversion to this type will be performed on the
                    server. If possible this will be determined automatically
                    using isinstance(), otherwise "auto" will be set:
                        auto (default) - no type conversion
                        string - `str`
                        boolean - `bool`
                        int - `int`
                        float - `float`
                "defaultValue" - the value that is considered the default,
                    either because a model instance will use this default value
                    if unspecified, or a client should present this option as
                    the default for a user
                "validators" - a `dict` mapping a validator name (e.g. "max")
                    and the value to be used in validation (e.g. 1000, for a
                    max value of 1000). This is intended to support client side
                    validation.
        """

        self.can_create = can_create
        self.can_delete = can_delete
        if no_update is not None and create is None:
            create = deepcopy(no_update)
        self.read = read or []
        self.no_read = no_read or []
        self.update = update or []
        self.no_update = no_update or [x for x in self.no_read if x not in self.update]
        self.create = create or []
        self.no_create = no_create or [x for x in self.no_update if x not in self.create]

        self.no_read.extend(self.never_read)
        self.no_update.extend(self.never_update)

        self.data_spec = data_spec or {}

    def __call__(self, cls):

        def _crud_perms(cls):
            crud_perms = {
                'can_create': self.can_create,
                'can_delete': self.can_delete,
                'read': [],
                'update': [],
                'create': []
            }

            read = self.read
            for name in collect_superclass_attr_names(cls):
                if not name.startswith('_'):
                    attr = getattr(cls, name)
                    properties = (InstrumentedAttribute, property, ClauseElement)
                    primitives = (int, float, bool, datetime, date, time, six.binary_type, six.text_type, uuid.UUID)
                    if isinstance(attr, properties) or isinstance(attr, primitives):
                        read.append(name)

            read = list(set(read))
            for name in read:
                if not self.no_read or name not in self.no_read:
                    crud_perms['read'].append(name)

            update = self.update + deepcopy(crud_perms['read'])
            update = list(set(update))
            for name in filter(lambda s: s not in self.no_update, update):
                if name in cls.__table__.columns:
                    crud_perms['update'].append(name)
                else:
                    attr = getattr(cls, name)
                    if (isinstance(attr, property) and getattr(attr, 'fset', False)) or (
                            isinstance(attr, InstrumentedAttribute)
                            and isinstance(attr.property, RelationshipProperty)
                            and attr.property.viewonly is not True):
                        crud_perms['update'].append(name)

            create = self.create + deepcopy(crud_perms['update'])
            for name in self.always_create:
                create.append(name)
                if name in self.no_create:
                    self.no_create.remove(name)

            create = list(set(create))
            for name in create:
                if not self.no_create or name not in self.no_create:
                    crud_perms['create'].append(name)

            cls._cached_crud_perms = crud_perms
            return cls._cached_crud_perms

        def _crud_spec(cls):
            crud_perms = cls._crud_perms

            field_names = list(
                set(crud_perms['read']) |
                set(crud_perms['update']) |
                set(crud_perms['create']) |
                set(self.data_spec.keys()))

            fields = {}
            for name in field_names:
                validators = getattr(cls, '_validators', {}).get(name, [])
                # If using different validation decorators or in the data spec
                # causes multiple spec kwargs to be specified, we're going to
                # error here for duplicate keys in dictionaries. Since we don't
                # want to allow two different expected values for maxLength
                # being sent in a crud spec for example.
                field_validator_kwargs = {
                    key: value
                    # Collect each spec_kwarg for all validators of an attr.
                    for validator in validators
                    for key, value in validator.get('spec_kwargs', {}).items()
                }

                if field_validator_kwargs:
                    self.data_spec.setdefault(name, {})
                    # Manually specified crud validator keyword arguments
                    # overwrite the decorator-supplied keyword arguments.
                    field_validator_kwargs.update(self.data_spec[name].get('validators', {}))
                    self.data_spec[name]['validators'] = field_validator_kwargs

                field = deepcopy(self.data_spec.get(name, {}))
                field['name'] = name
                try:
                    attr = getattr(cls, name)
                except AttributeError:
                    # If the object doesn't have the attribute, AND it's in the
                    # field list, that means we're assuming it was manually
                    # specified in the data_spec argument.
                    fields[name] = field
                    continue

                field['read'] = name in crud_perms['read']
                field['update'] = name in crud_perms['update']
                field['create'] = name in crud_perms['create']

                if field['read'] or field['update'] or field['create']:
                    fields[name] = field
                elif name in fields:
                    del fields[name]
                    continue

                if 'desc' not in field and not is_data(attr):
                    # No desc specified, but there's a relevant docstring,
                    # so we use that instead.

                    # If there's 2 consecutive newlines, assume that there's a
                    # separator in the docstring and that the top part only
                    # is the description, if there's not, use the whole thing.
                    # Either way, replace newlines with spaces since docstrings
                    # often break the same sentence accross lines due to space.
                    doc = inspect.getdoc(attr)
                    if doc:
                        doc = doc.partition('\n\n')[0].replace('\n', ' ').strip()
                        field['desc'] = doc

                if 'type' not in field:
                    if isinstance(attr, InstrumentedAttribute) and isinstance(attr.property, ColumnProperty):
                        col = attr.property.columns[0]
                        col_type = type(col.type)
                        field['type'] = cls._type_map.get(col_type, 'auto')
                        field_default = getattr(col, 'default', None)
                        # Only put the default here if it exists, and it's not
                        # an automatic thing like "time.utcnow()".
                        if field_default is not None \
                                and field['type'] != 'auto' \
                                and not isinstance(field_default.arg, (Callable, property)):
                            field['defaultValue'] = field_default.arg
                    elif hasattr(attr, "default"):
                        field['defaultValue'] = attr.default
                    else:
                        field['type'] = cls._type_map.get(type(attr), 'auto')
                        # Only set a default if this isn't a property or some
                        # other kind of "constructed attribute".
                        if field['type'] != 'auto' and not isinstance(attr, (Callable, property)):
                            field['defaultValue'] = attr
                if isinstance(attr, InstrumentedAttribute) and isinstance(attr.property, RelationshipProperty):
                    field['_model'] = attr.property.mapper.class_.__name__

            crud_spec = {'fields': fields}
            cls._cached_crud_spec = crud_spec
            return cls._cached_crud_spec

        def _type_map(cls):
            return dict(cls.type_map_defaults, **cls.type_map)

        cls._type_map = cached_classproperty(_type_map)
        cls._crud_spec = cached_classproperty(_crud_spec)
        cls._crud_perms = cached_classproperty(_crud_perms)
        return cls


class crud_validation(object):
    """
    Base class for adding validators to a model.

    Supports adding to the crud spec, or to the save action.
    """

    def __init__(self, attr_name, model_validator, validator_message, **spec_kwargs):
        """

        Args:
            attr_name (str): The attribute to which this validator applies.
            model_validator (callable): A callable that accepts the attribute
                value and returns False or None if invalid, or True if the
                value is valid.
            validator_message (str): Failure message if the validation fails.
            **spec_kwargs: The key/value pairs that should be added to the
                the crud spec for this attribute name. This generally supports
                making the same sorts of validations in a client (e.g.
                javascript).
        """
        self.attr_name = attr_name
        self.model_validator = model_validator
        self.validator_message = validator_message
        self.spec_kwargs = spec_kwargs

    def __call__(self, cls):
        if not hasattr(cls, '_validators'):
            cls._validators = {}
        else:
            # In case we subclass something with a _validators attribute.
            cls._validators = deepcopy(cls._validators)

        cls._validators.setdefault(self.attr_name, []).append({
            'model_validator': self.model_validator,
            'validator_message': self.validator_message,
            'spec_kwargs': self.spec_kwargs
        })
        return cls


class text_length_validation(crud_validation):
    """
    Validates the length of a text attribute.
    """
    def __init__(
            self, attr_name, min_length=None, max_length=None,
            min_text='The minimum length of this field is {0}.',
            max_text='The maximum length of this field is {0}.',
            allow_none=True):

        def model_validator(instance, text):
            if text is None:
                return allow_none
            text_length = len(six.text_type(text))
            return all([min_length is None or text_length >= min_length,
                        max_length is None or text_length <= max_length])

        kwargs = {}
        if min_length is not None:
            kwargs['minLength'] = min_length
            if max_text is not None:
                kwargs['minLengthText'] = min_text
        if max_length is not None:
            kwargs['maxLength'] = max_length
            if max_text is not None:
                kwargs['maxLengthText'] = max_text

        message = 'Length of value should be between {} and {} (inclusive; ' \
            'None means no min/max).'.format(min_length, max_length)

        crud_validation.__init__(self, attr_name, model_validator, message, **kwargs)


class regex_validation(crud_validation):
    """
    Validates a text field against the given regular expression.
    """
    def __init__(self, attr_name, regex, message):

        def regex_validator(instance, text):
            # If the field isn't nullable, that will trigger an error later
            # at the SQLAlchemy level, but since None can't be passed to
            # re.search() we want to pass this validation check.
            if text is None:
                return True

            return re.search(regex, text) is not None

        crud_validation.__init__(
            self,
            attr_name,
            regex_validator,
            message,
            regexText=message,
            regexString=regex)
